/*
 * Copyright (c) 1992, 1996, 1998, 2002, 2009, 2010
 *	Craig Leres
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that: (1) source code
 * distributions retain the above copyright notice and this paragraph
 * in its entirety, and (2) distributions including binary code include
 * the above copyright notice and this paragraph in its entirety in
 * the documentation or other materials provided with the distribution
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 */
#ifndef lint
static const char copyright[] =
    "@(#) Copyright (c) 1992, 1996, 1998, 2002, 2009, 2010\nCraig Leres\n";
static const char rcsid[] =
    "@(#) $Id: captive.c 37 2010-10-15 00:03:44Z leres $ (XSE)";
#endif

#include <sys/param.h>
#ifdef HAVE_JAIL
#include <sys/jail.h>
#endif

#include <errno.h>
#include <pwd.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <syslog.h>
#include <unistd.h>

/* Forwards */
const char *getjailuser(const char *);
int main(int, char **);

/* Determine jail user */
const char *
getjailuser(const char *dir)
{
	char *cp;
	struct passwd *pwd;
	char jaildir[PATH_MAX];
	static char jailuser[128];

	/* The captive user's home directory should be inside the jail user */
	realpath(dir, jaildir);
	cp = jaildir + strlen(jaildir) - strlen(dir);
	if (cp <= jaildir) {
		syslog(LOG_ERR, "getjailuser(%s): Bad captive home dir\n", dir);
		return (NULL);
	}
	*cp = '\0';

	setpwent();
	while ((pwd = getpwent()) != NULL) {
		if (strcmp(jaildir, pwd->pw_dir) == 0) {
			/* Found it! */
			(void)strncpy(jailuser, pwd->pw_name, sizeof(jailuser));
			jailuser[sizeof(jailuser) - 1] = '\0';
			endpwent();
			return (jailuser);
		}
	}
	endpwent();
	return (NULL);
}

int
main(int argc, char **argv)
{
	uid_t uid, euid;
	struct passwd *pwd;
	char *csh = "/bin/csh";
	char *name = "-csh";
	char root[PATH_MAX];
	char dir[PATH_MAX];
	const char *jailuser;
#ifdef HAVE_JAIL
	char hostname[MAXHOSTNAMELEN];
	struct jail jailinfo;
#endif

	openlog("captive", 0, LOG_AUTH);

	/* Get the captive user's info */
	uid = getuid();
	if (uid == 0) {
		syslog(LOG_ERR, "uid == 0");
		exit(1);
	}
	euid = geteuid();
	if (euid != 0) {
		syslog(LOG_ERR, "euid != 0 (%d)", euid);
		exit(1);
	}

	/* Get the captive user's home directory */
	pwd = getpwuid(uid);
	if (pwd == NULL) {
		syslog(LOG_ERR, "No such user with uid %d", uid);
		exit(1);

	}

	/* The captive user's home directory might be a symlink */
	strncpy(dir, pwd->pw_dir, sizeof(dir));
	dir[sizeof(dir) - 1] = '\0';
	if ((jailuser = getjailuser(dir)) == NULL &&
		(readlink(dir, dir, sizeof(dir)) < 0 ||
		(jailuser = getjailuser(dir)) == NULL)) {
		syslog(LOG_ERR, "Can't find jail user for captive uid %d", uid);
		exit(1);
	}

	/* Get the jail root directory (jail user's home directory) */
	pwd = getpwnam(jailuser);
	if (pwd == NULL) {
		syslog(LOG_ERR, "Can't find %s account", jailuser);
		exit(1);
	}
	(void)strncpy(root, pwd->pw_dir, sizeof(root));
	root[sizeof(root) - 1] = '\0';

	/* Now get the captive user's info again */
	pwd = getpwuid(uid);
	if (pwd == NULL) {
		syslog(LOG_ERR, "No such user with uid %d", uid);
		exit(1);
	}

	syslog(LOG_INFO, "%s inside %s", pwd->pw_name, jailuser);

#ifdef HAVE_JAIL
	memset(&jailinfo, 0, sizeof(jailinfo));
#ifdef JAIL_API_VERSION
	jailinfo.version = JAIL_API_VERSION;
#endif
	jailinfo.path = root;
	if (gethostname(hostname, sizeof(hostname)) < 0) {
		syslog(LOG_ERR, "gethostname: %m");
		exit(1);
	}
	jailinfo.hostname = hostname;
	if (jail(&jailinfo) < 0) {
		syslog(LOG_ERR, "jail %s: %m", root);
		exit(1);
	}
#else
	if (chroot(root) < 0) {
		syslog(LOG_ERR, "chroot %s: %m", root);
		exit(1);
	}
	if (chdir(pwd->pw_dir) < 0) {
		syslog(LOG_ERR, "chdir %s: %m", pwd->pw_dir);
		exit(1);
	}
#endif
	if (setgid(pwd->pw_gid) < 0) {
		syslog(LOG_ERR, "setgid %d: %m", pwd->pw_gid);
		exit(1);
	}
	if (setuid(pwd->pw_uid) < 0) {
		syslog(LOG_ERR, "setuid %d: %m", pwd->pw_uid);
		exit(1);
	}

	/* Remove leading dash if not interactive (rshd doesn't set TERM) */
	if (getenv("TERM") == NULL)
		++name;
	argv[0] = name;
	execv(csh, argv);

	printf("execv failed: %s\n", strerror(errno));
	syslog(LOG_ERR, "exec \"%s\": %m", csh);
	exit(1);
}
