#!/bin/sh
# @(#) $Id: updatecaptive.sh 32 2010-10-14 20:49:09Z leres $ (LBL)
#
# updatecaptive.sh - update captive chroot bins and libs
#
# Sends an update report to stdout if changes were made.
#
# To configure, add variables to /etc/rc.conf:
#
#	captive_enable="YES"
#	captive_users		# optional captive accounts (default: captive)
#	captive*_files		# files to copy into jail
#	captive*_bins		# binaries (automatic library dependancies)
#	captive*_libs		# shared libraries (automatic dependancies)
#

umask 022

prog=`basename $0`
t1=/tmp/${prog}.1.$$

trap 'rm -f ${t1}; exit 1' 1 2 3 15

PATH="${PATH}:/usr/local/bin"
export PATH

usage() {
	echo "usage: ${prog} [-n]" 1>&2
	exit 1
}

args=`getopt n $*`
if [ $? -ne 0 ]; then
	usage
fi

set -- ${args}

dryrun=0

for a do
        case "${a}" in
 
        -n)
                dryrun=1
                shift
                ;;

        --)
                shift
                break
                ;;
        esac
done

if [ $# -ne 0 ]; then
        usage
fi

createdir() {
	local dir mode
	if [ $# -ne 2 ]; then
		echo "${prog}: createdir: bad args" 1>&2
		exit 1
	fi
	dir=$1
	mode=$2

	if [ ! -d ${dir} ]; then
		cmd="mkdir -p ${dir}"
		echo "+ ${cmd}"
		if [ ${dryrun} -eq 0 ]; then
			${cmd} || exit 1
		fi

		cmd="chown 0:0 ${dir}"
		echo "+ ${cmd}"
		if [ ${dryrun} -eq 0 ]; then
			${cmd} || exit 1
		fi

		cmd="chmod ${mode} ${dir}"
		echo "+ ${cmd}"
		if [ ${dryrun} -eq 0 ]; then
			${cmd} || exit 1
		fi

		cmd="ls -ld ${dir}"
		echo "+ ${cmd}"
		if [ ${dryrun} -eq 0 ]; then
			${cmd} || exit 1
		fi
	fi
}

# Define replacement shell function if running on linux
realpath /dev/null >/dev/null 2>&1
if [ $? -ne 0 ]; then
	realpath() {
		if [ ! -h $1 ]; then
			echo $1
			return
		fi
		dest="`readlink $1`"
		case "${dest}" in

		/*)
			echo ${dest}
			;;
		*)
			dir="`dirname $1`"
			echo "${dir}/${dest}"
		esac
	}
fi

. /etc/rc.conf

enable="${captive_enable:-NO}"

if [ ${enable} != "YES" ]; then
	echo "${prog}: captive_enable is not enabled in /etc/rc.conf" 1>&2
	exit 1
fi

# Allow multiple captive users; default to "captive"
users="${captive_users:-captive}"

for user in ${users}; do
	# Files
	files="/etc/localtime"
	for file in /etc/termcap /usr/share/misc/termcap \
	    /usr/share/misc/termcap.db ; do
		if [ -r ${file} ]; then
			files="${files} ${file}"
		fi
	done
	eval tfiles="\${${user}_files}"
	files="${files} ${tfiles}"

	# Bins
	eval tbins="\${${user}_bins}"
	bins="/bin/csh ${tbins}"

	# Libs
	libs=""
	for file in /libexec/ld-* /lib/ld-*; do
		if [ -r ${file} ]; then
			libs="${libs} ${file}"
		fi
	done
	eval tlibs="\${${user}_libs}"
	libs="${libs} ${tlibs}"

	# Generate list of shared libraries
	rm -f ${t1}
	for file in ${bins}; do
		if [ ! -x ${file} ]; then
			continue
		fi
		ldd ${file} 2>&1 |
		    awk '/not a dynamic executable/ {
			    exit
		    }
		    $2 == "=>" {
			    print $3
		    }' >> ${t1}
	done
	sharedlibs="`sort -u ${t1}`"

	# Resolve symlinks
	for file in ${sharedlibs}; do
		realpath ${file} >> ${t1}
	done
	sharedlibs="`sort -u ${t1}`"
	rm ${t1}

	# Get rid of duplicates
	files="`echo ${files} | fmt -1 | sort -u`"
	bins="`echo ${bins} | fmt -1 | sort -u`"
	libs="`echo ${libs} | fmt -1 | sort -u`"
	sharedlibs="`echo ${sharedlibs} | fmt -1 | sort -u`"

	# Create directories
	eval dir=~${user}
	createdir ${dir} 755
	createdir ${dir}/dev 555
	createdir ${dir}/tmp 1777

	createdir ${dir}/var 755
	createdir ${dir}/var/log 755
	createdir ${dir}/var/spool 755
	createdir ${dir}/var/spool/lock 1777
	createdir ${dir}/var/tmp 1777
	createdir ${dir}/var/tmp/vi.recover 1777

	# Update files, bins, libs and sharedlibs
	rsync="rsync -aiHRO --delete $*"
	if [ ${dryrun} -ne 0 ]; then
		rsync="${rsync} -n"
	fi
	${rsync} ${files} ${bins} ${libs} ${sharedlibs} ${dir} > ${t1} 2>&1
	if [ $? -ne 0 ]; then
		echo "+ ${rsync} [files ...] ${dir}
		cat ${t1}
		rm ${t1}
		exit 1
	fi

	# Only print if we updated files
	if [ -s ${t1} ]; then
		echo "+ ${rsync} [files ...] ${dir}
		cat ${t1}
	fi
done

rm ${t1}
